#  ***** GPL LICENSE BLOCK *****
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
#  All rights reserved.
#  ***** GPL LICENSE BLOCK *****


import bpy, bmesh
import struct
import os
from collections import namedtuple


# type definitions
BSPHeader = namedtuple('BSPHeader', 
    ("version,"
    "entities_ofs,  entities_size,"
    "planes_ofs,    planes_size,"
    "miptex_ofs,    miptex_size,"
    "verts_ofs,     verts_size,"
    "visilist_ofs,  visilist_size,"
    "nodes_ofs,     nodes_size,"
    "texinfo_ofs,   texinfo_size,"
    "faces_ofs,     faces_size,"
    "lightmaps_ofs, lightmaps_size,"
    "clipnodes_ofs, clipnodes_size,"
    "leaves_ofs,    leaves_size,"
    "lface_ofs,     lface_size,"
    "edges_ofs,     edges_size,"
    "ledges_ofs,    ledges_size,"
    "models_ofs,    models_size")
    )
fmt_BSPHeader = '<31i'

BSPModel = namedtuple('BSPModel',
    ("bbox_min_x,  bbox_min_y,  bbox_min_z,"
    "bbox_max_x,   bbox_max_y,  bbox_max_z,"
    "origin_x,     origin_y,    origin_z,"
    "node_id0, node_id1, node_id2, node_id3,"
    "numleafs,"
    "face_id,"
    "face_num")
    )
fmt_BSPModel = '<9f7i'

BSPFace = namedtuple('BSPFace',
    ("plane_id,"
    "size,"
    "ledge_id,"
    "ledge_num,"
    "texinfo_id,"
    "lighttype,"
    "lightlevel,"
    "light0, light1,"
    "lightmap")
    )
fmt_BSPFace = '<HHiHHBBBBi'

BSPVertex = namedtuple('BSPVertex', 'x, y, z')
fmt_BSPVertex = '<fff'

BSPEdge = namedtuple('BSPEdge', 'vertex0, vertex1')
fmt_BSPEdge = '<HH'

BSPTexInfo = namedtuple('BSPTexInfo',
    ("s_x,  s_y,    s_z,    s_dist,"
    "t_x,   t_y,    t_z,    t_dist,"
    "texture_id,"
    "animated")
    )
fmt_BSPTexInfo = '<8f2I'

BSPMipTex = namedtuple('BSPMipTex', 'name, width, height, ofs1, ofs2, ofs4, ofs8')
fmt_BSPMipTex = '<16s6I'


# functions
def load_palette(filepath, pixel_type):
    with open(filepath, 'rb') as file:
        colors_byte = struct.unpack('<768B', file.read(768))
        if(pixel_type == 'BYTE'):
            return colors_byte
        else:
            colors_float = [float(c)/255.0 for c in colors_byte]
            return colors_float

def load_textures(context, filepath):
    with open(filepath, 'rb') as file:
        # read file header
        header_data = file.read(struct.calcsize(fmt_BSPHeader))
        header = BSPHeader._make(struct.unpack(fmt_BSPHeader, header_data))

        # get the list of miptex in the miptex lump (basically a simplified .WAD file inside the bsp)
        file.seek(header.miptex_ofs)
        num_miptex = struct.unpack('<i', file.read(4))[0]
        miptex_ofs_list = struct.unpack('<%di' % num_miptex, file.read(4*num_miptex))

        # load the palette colours (will be converted to RGB float format)
        script_path = os.path.dirname(os.path.abspath(__file__)) + "/"
        colors = load_palette(script_path + "palette.lmp", "FLOAT")

        # return a list of texture information and image data
        # entry format: dict(name, width, height, image)
        texture_data = []

        # load each mip texture
        for miptex_id in range(num_miptex):
            ofs = miptex_ofs_list[miptex_id]
            # get the miptex header
            file.seek(header.miptex_ofs + ofs)
            miptex_data = file.read(struct.calcsize(fmt_BSPMipTex))
            miptex = BSPMipTex._make(struct.unpack(fmt_BSPMipTex, miptex_data))
            miptex_size = miptex.width * miptex.height
            miptex_name = miptex.name.decode('ascii').rstrip('\0')
            print("[%d] \'%s\' (%dx%d %dbytes)" % (miptex_id, miptex_name, miptex.width, miptex.height, miptex_size))

            # get the paletized image pixels
            file.seek(header.miptex_ofs + ofs + miptex.ofs1)
            pixels_pal = struct.unpack('<%dB' % miptex_size, file.read(miptex_size))

            # convert the paletized pixels into regular rgba pixels
            # note that i is fiddled with in order to reverse Y
            pixels = []
            for y in reversed(range(miptex.height)):
                i = miptex.width * y
                for x in range(miptex.width):
                    c = pixels_pal[i+x] * 3
                    pixels.append(colors[c])    # red
                    pixels.append(colors[c+1])  # green
                    pixels.append(colors[c+2])  # blue
                    pixels.append(1.0)          # alpha
                    
            # create an image and save it
            image = bpy.data.images.new(miptex_name, width=miptex.width, height=miptex.height)
            image.pixels = pixels
            texture_data.append(dict(name=miptex_name, width=miptex.width, height=miptex.height, image=image))
            # image.filepath_raw = "/tmp/%s.png" % miptex_name
            # image.file_format = 'PNG'
            # image.save()

        # used by the import_bsp function to set texture mapping for faces
        # and optionally to create materials in the scene
        return texture_data

def mesh_add(mesh_id):
    if bpy.context.scene.objects.active:
        bpy.ops.object.mode_set(mode='OBJECT')
    bpy.ops.object.add(type='MESH', enter_editmode=True)
    obj = bpy.context.object
    obj.name = "bsp_model_%d" % mesh_id
    obj.show_name = True
    obj_mesh = obj.data
    obj_mesh.name = obj.name + "_mesh"
    return obj

def mesh_append_face(obj, bm, vertex_list, uv_list, texture_name):
    # add vertices and uvs before creating the new face
    # note that there is a little faff to get the face normals and uvs in the correct order
    for vert in vertex_list:
        bm.verts.new( vert )
    bm.faces.new((bm.verts[i] for i in reversed(range(-len(vertex_list), 0))))

    # add uvs to the new face
    uv_layer = bm.loops.layers.uv.verify()
    bm.faces.layers.tex.verify()  # currently blender needs both layers.

    face = bm.faces[-1]
    for i, loop in enumerate(face.loops):
        uv = loop[uv_layer].uv
        uv[0] = uv_list[-i-1][0]
        uv[1] = -uv_list[-i-1][1]
        # print("UV%d: %.2f, %.2f" % (i, uv[0], uv[1]))

    # early escape if not including materials
    if texture_name == "":
        return

    # assign material
    try:
        material_names = [ m.name for m in obj.data.materials ]
        material_id = material_names.index(texture_name)
    except: #ValueError:
        obj.data.materials.append(bpy.data.materials[texture_name])
        material_id = len(obj.data.materials) - 1

    face.material_index = material_id

def calculate_uv(vertex, texinfo, texture_data):
    # u = (dot(vtx, s) + s_dist) / texture_width
    # v = (dot(vtx, t) + t_dist) / texture_height
    u = (vertex[0] * texinfo.s_x + vertex[1] * texinfo.s_y + vertex[2] * texinfo.s_z) + texinfo.s_dist
    v = (vertex[0] * texinfo.t_x + vertex[1] * texinfo.t_y + vertex[2] * texinfo.t_z) + texinfo.t_dist
    u /= texture_data['width']
    v /= texture_data['height']
    return [u, v]

def create_materials(texture_data):
    for texture in texture_data:
        name = texture['name']
        image = texture['image']

        # pack image data in .blend
        image.pack(as_png = True)
        # create texture from image
        texture = bpy.data.textures.new(name, type='IMAGE')
        texture.image = image
        texture.use_alpha = False
        # create material
        mat = bpy.data.materials.new(name)
        mat.diffuse_intensity = 1.0
        mat.specular_intensity = 0.0
        mat.preview_render_type = 'CUBE'
        # assign texture to material
        mtex = mat.texture_slots.add()
        mtex.texture = texture
        mtex.texture_coords = 'UV'

def import_bsp(context, filepath, options):
    # TODO: Clean this up; Perhaps by loading everything into lists to begin with
    with open(filepath, 'rb') as file:
        header_data = file.read(struct.calcsize(fmt_BSPHeader))
        header = BSPHeader._make(struct.unpack(fmt_BSPHeader, header_data))

        num_models = int(header.models_size / struct.calcsize(fmt_BSPModel))
        num_faces = int(header.faces_size / struct.calcsize(fmt_BSPFace))
        num_edges = int(header.edges_size / struct.calcsize(fmt_BSPEdge))
        num_verts = int(header.verts_size / struct.calcsize(fmt_BSPVertex))

        print("-- IMPORTING BSP --")
        print("Source file: %s" % filepath)
        print("bsp contains %d models. (faces = %d, edges = %d, verts = %d)" % (num_models, num_faces, num_edges, num_verts))

        # read models, faces, edges and vertices into buffers
        file.seek(header.models_ofs)
        model_data = file.read(header.models_size)
        file.seek(header.faces_ofs)
        face_data = file.read(header.faces_size)
        file.seek(header.edges_ofs) # actual edges
        edge_data = file.read(header.edges_size)
        file.seek(header.texinfo_ofs)
        texinfo_data = file.read(header.texinfo_size)

        # read in the list of edges and store in readable form (flat list of ints)
        file.seek(header.ledges_ofs)
        edge_index_list = struct.unpack('<%di' % int(header.ledges_size/4), file.read(header.ledges_size))
        # do the same with vertices (flat list of floats)
        file.seek(header.verts_ofs)
        vertex_list = struct.unpack('<%df' % int(header.verts_size/4), file.read(header.verts_size))

    # load texture data (name, width, height, image)
    print("-- LOADING TEXTURES --")
    texture_data = load_textures(context, filepath)
    if options['create_materials']:
        create_materials(texture_data)

    # create some structs for storing data
    model_size = struct.calcsize(fmt_BSPModel)
    model_struct = struct.Struct(fmt_BSPModel)
    face_size = struct.calcsize(fmt_BSPFace)
    face_struct = struct.Struct(fmt_BSPFace)
    edge_size = struct.calcsize(fmt_BSPEdge)
    edge_struct = struct.Struct(fmt_BSPEdge)
    texinfo_size = struct.calcsize(fmt_BSPTexInfo)
    texinfo_struct = struct.Struct(fmt_BSPTexInfo)

    print("-- LOADING MODELS --")
    start_model = 0
    if options['worldspawn_only'] == True:
        end_model = 1
    else:
        end_model = num_models

    for m in range(start_model, end_model):
        model_ofs = m * model_size
        model = BSPModel._make(model_struct.unpack_from(model_data[model_ofs:model_ofs+model_size]))
        print("[%d] %d faces" % (m, model.face_num))

        # create new mesh
        obj = mesh_add(m)
        bm = bmesh.from_edit_mesh(obj.data)

        for f in range(0, model.face_num):
            face_ofs = (model.face_id + f) * face_size
            face = BSPFace._make(face_struct.unpack_from(face_data[face_ofs:face_ofs+face_size]))
            texinfo_ofs = face.texinfo_id * texinfo_size
            texinfo = BSPTexInfo._make(texinfo_struct.unpack_from(texinfo_data[texinfo_ofs:texinfo_ofs+texinfo_size]))

            # populate a list with vertices
            face_vertices = []
            face_uvs = []

            for i in range(0,face.ledge_num):
                edge_index = edge_index_list[face.ledge_id+i]
                # assuming vertex order is 0->1 
                edge_ofs = edge_index * edge_size
                vert_id = 0
                if edge_index < 0:
                    # vertex order is 1->0
                    edge_ofs = -edge_index * edge_size
                    vert_id = 1
                
                edge = BSPEdge._make(edge_struct.unpack_from(edge_data[edge_ofs:edge_ofs+edge_size]))
                vofs = edge[vert_id] * 3

                vertex = [vertex_list[vofs], vertex_list[vofs+1], vertex_list[vofs+2]]
                uv = calculate_uv(vertex, texinfo, texture_data[texinfo.texture_id])

                scale = options['scale']
                vertex[0] *= scale
                vertex[1] *= scale
                vertex[2] *= scale

                face_vertices.append(vertex)
                face_uvs.append(uv)

            # append face to mesh
            if options['create_materials']:
                texture_name = texture_data[texinfo.texture_id]['name']
            else:
                texture_name = ""
            mesh_append_face(obj, bm, face_vertices, face_uvs, texture_name)

        # update the mesh with data from the bmesh
        bmesh.update_edit_mesh(obj.data)

    bpy.ops.object.mode_set(mode='OBJECT')
    print("-- IMPORT COMPLETE --")